/*
 *  $Id: mtrace.cpp,v 1.4 2011/09/10 23:54:49 amelinte Exp amelinte $
 *
 *  Copyright 2011 Aurelian Melinte. 
 *  Released under LGPL 3.0 or later. 
 *
 *  mtrace()/muntrace() & tools.  Based on glibc 2.6.1. 
 */

/* More debugging hooks for `malloc'.
   Copyright (C) 1991-1994,1996-2003, 2004 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
             Written April 2, 1991 by John Gilmore of Cygnus Support.
             Based on mcheck.c by Mike Haertel.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <malloc.h>
#include <mcheck.h>

#include <assert.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <execinfo.h> 
#include <pthread.h>

#include <signal.h>

#include <exception>
#include <iostream>
#include <cxxabi.h>


#define DEMANGLECXX

static FILE *_mallstream = NULL;
static const char _mallenv[]= "MALLOC_TRACE";
static char *_malloc_trace_buffer = NULL;

static pthread_mutex_t _hook_lock  = PTHREAD_MUTEX_INITIALIZER;


/* 
 * Demangling functions are calling malloc(). Do not trace these allocations
 * (infinite loop). 
 */
#ifndef __cplusplus
typedef enum {false, true} bool;
#endif
static __thread bool _in_trace = false;


/*
 *  Trace buffer.
 */
#define TRACE_ITEM_SIZE     (4*1024 - 1)
#define STACK_DEPTH         (40)
#define TRACE_BUFFER_SIZE   ((STACK_DEPTH*2*(TRACE_ITEM_SIZE+1))+1)
static __thread char   _trace_buf[TRACE_BUFFER_SIZE] = {0}; 
static __thread size_t _trace_buf_idx = 0;



/* Address to breakpoint on accesses to... */
__ptr_t mallwatch;


/* File name and line number information, for callers that had
   the foresight to call through a macro.  */
char *_mtrace_file;
int _mtrace_line;


/* Old hook values.  */
static void (*_tr_old_free_hook) (__ptr_t ptr, const __ptr_t);
static __ptr_t (*_tr_old_malloc_hook) (__malloc_size_t size, const __ptr_t);
static __ptr_t (*_tr_old_realloc_hook) (__ptr_t ptr, __malloc_size_t size,
                                        const __ptr_t);
static __ptr_t (*_tr_old_memalign_hook) (__malloc_size_t __alignment,
                                         __malloc_size_t __size,
                                         __const __ptr_t);

/* This function is called when the block being alloc'd, realloc'd, or
   freed has an address matching the variable "mallwatch".  In a debugger,
   set "mallwatch" to the address of interest, then put a breakpoint on
   tr_break.  */

extern "C"
void
tr_break ()
{
}


static void 
reset_trace_buffer(void)
{
    _trace_buf_idx = 0;
}

static void 
flush_trace_buffer(void)
{
    fprintf(_mallstream, "%s", _trace_buf); 
    fflush(_mallstream);
    _trace_buf_idx = 0;
}


inline const char*
demangle(const char* symname, char* buf, size_t buflen)
{
#ifdef __cplusplus
    if (symname && *symname) {
        int status = 0;
        /* Note: if the demangle 'buf' we pass in now is not big enough, 
         * we will crash in free(). */
        abi::__cxa_demangle(symname, buf, &buflen, &status);
        if (0 == status) {
            return buf;
        }
    }

    return symname;
#else
    return symname;
#endif
}

inline static void 
print_address(const __ptr_t addr)
{
    /* Note: dladdr() locks the DL mutex. */

    Dl_info info;
    if (dladdr (addr, &info)) {
        char buf[2*TRACE_ITEM_SIZE+1] = {0};
        char symbuf[TRACE_ITEM_SIZE+1] = {0};
        
        const char* sname = info.dli_sname != NULL ? info.dli_sname : "";
        const char* fname = info.dli_fname != NULL ? info.dli_fname : "";
        bool   positive = (addr >= (const __ptr_t) info.dli_saddr);
        long   delta    = labs((char*)addr - (char*) info.dli_saddr);

        snprintf(buf, 2*TRACE_ITEM_SIZE, "[%p] %s (%s%c%0lx)", 
                 addr,
                 demangle(sname, symbuf, TRACE_ITEM_SIZE), fname, 
                 positive ? '+' : '-',
                 delta);
        
        _trace_buf_idx += snprintf(_trace_buf+_trace_buf_idx, TRACE_BUFFER_SIZE-_trace_buf_idx, "%s\n", buf);
    }
}


inline static void
print_stack(void **buffer, int size)
{
#if 0 /* Names are not demangled */
    backtrace_symbols_fd(buffer, size, fileno(_mallstream));
#else
    for (int i = 0; i< size; i++) {
        print_address((const __ptr_t)buffer[i]);
    }
#endif
}

inline static int 
mbacktrace(void **buffer, int size)
{
    return backtrace(buffer, size);
}


static void
dump_stack()
{
    void *stack_buf[STACK_DEPTH] = {0};
    int  n = 0;
    
    n = mbacktrace(stack_buf, STACK_DEPTH);
    
    _trace_buf_idx += snprintf(_trace_buf+_trace_buf_idx, TRACE_BUFFER_SIZE-_trace_buf_idx, "**[ Stack: %u\n", n);
    print_stack(stack_buf, n);
    _trace_buf_idx += snprintf(_trace_buf+_trace_buf_idx, TRACE_BUFFER_SIZE-_trace_buf_idx, "**] Stack\n\n"); 
    
    flush_trace_buffer();
}


static void
trace_where (const __ptr_t caller)
{
  assert(_in_trace == true);
  reset_trace_buffer();
  
  if (_mtrace_file)
  {
      _trace_buf_idx += snprintf (_trace_buf+_trace_buf_idx, TRACE_BUFFER_SIZE-_trace_buf_idx, "@ %s:%d ", _mtrace_file, _mtrace_line);
      _mtrace_file = NULL;
  }
  else
  {
    if (caller != NULL)
    {
      Dl_info info;
      if (dladdr (caller, &info)) // link with -ldl 
      {
        char buf[TRACE_ITEM_SIZE+1] = {0};
        if (info.dli_sname != NULL)
        {
            //char symbuf[TRACE_ITEM_SIZE+1] = {0};
            const char* sname = info.dli_sname; //demangle(info.dli_sname, symbuf, TRACE_ITEM_SIZE);
            
            bool   positive = (caller >= (const __ptr_t) info.dli_saddr);
            long   delta    = labs((char*)caller - (char*) info.dli_saddr);
            
            snprintf(buf, TRACE_ITEM_SIZE, "(%s%c%0lx)", 
                     sname, //symbuf, 
                     positive ? '+' : '-',
                     delta);
        }

        _trace_buf_idx += snprintf(_trace_buf+_trace_buf_idx, TRACE_BUFFER_SIZE-_trace_buf_idx, 
                                   "@ %s%s%s[%p] ",
                                   info.dli_fname ?: "", 
                                   info.dli_fname ? ":" : "",
                                   buf, 
                                   caller);
      }
      else
      {
          _trace_buf_idx += snprintf(_trace_buf+_trace_buf_idx, TRACE_BUFFER_SIZE-_trace_buf_idx, "@ [%p] ", caller);
      }
    }
  }
}

static inline void 
trace_what(const char * format,
           __ptr_t ptr,
           unsigned long long size)
{
    _trace_buf_idx += snprintf(_trace_buf+_trace_buf_idx, TRACE_BUFFER_SIZE-_trace_buf_idx, format, ptr, size);
}

static void
tr_freehook (__ptr_t ptr,
     const __ptr_t caller)
{
    if (ptr == NULL)
        return;
    
      /* The alloc/free operation is a result of the API calls we make to resolve
     * symbols while in a mem trace on this thread.  Do not log this.
        */
    bool oldval = __sync_val_compare_and_swap(&_in_trace, false, true);
    if (false == oldval)
    {
        assert(oldval == false && _in_trace == true);
        
        trace_where (caller);
        /* Be sure to print it first.  */
        trace_what ("- %p\n", ptr, 0/*ignored*/);
        dump_stack();
        
        oldval = __sync_val_compare_and_swap(&_in_trace, true, false);
        assert(oldval == true && _in_trace == false);
    }

    if (ptr == mallwatch)
        tr_break();

    {
        pthread_mutex_lock (&_hook_lock);
        __free_hook = _tr_old_free_hook;
        if (_tr_old_free_hook != NULL) {
            (*_tr_old_free_hook) (ptr, caller);
        } else {
            free (ptr);
        }
        __free_hook = tr_freehook;
        pthread_mutex_unlock(&_hook_lock);
    }
}

static __ptr_t
tr_mallochook (__malloc_size_t size,
     const __ptr_t caller)
{
    __ptr_t hdr = 0;

    {
        pthread_mutex_lock (&_hook_lock);

        __malloc_hook = _tr_old_malloc_hook;
        if (_tr_old_malloc_hook != NULL) {
            hdr = (__ptr_t) (*_tr_old_malloc_hook) (size, caller);
        } else {
            hdr = (__ptr_t) malloc (size);
        }
        __malloc_hook = tr_mallochook;

        pthread_mutex_unlock(&_hook_lock);
    }

    bool oldval = __sync_val_compare_and_swap(&_in_trace, false, true);
    if (false == oldval)
    {
        assert(oldval == false && _in_trace == true);
        
        trace_where (caller);
        /* We could be printing a NULL here; that's OK.  */
        trace_what ("+ %p %#lx\n", hdr, (unsigned long int) size);
        dump_stack();
        
        oldval = __sync_val_compare_and_swap(&_in_trace, true, false);
        assert(oldval == true && _in_trace == false);
    }

    if (hdr == mallwatch)
        tr_break ();

    return hdr;
}

static __ptr_t
tr_reallochook (
     __ptr_t ptr,
     __malloc_size_t size,
     const __ptr_t caller)
{
    __ptr_t hdr;

    if (ptr == mallwatch)
        tr_break ();

    {
        pthread_mutex_lock (&_hook_lock);

        __free_hook = _tr_old_free_hook;
        __malloc_hook = _tr_old_malloc_hook;
        __realloc_hook = _tr_old_realloc_hook;
        if (_tr_old_realloc_hook != NULL) {
            hdr = (__ptr_t) (*_tr_old_realloc_hook) (ptr, size, caller);
        } else {
            hdr = (__ptr_t) realloc (ptr, size);
        }
        __free_hook = tr_freehook;
        __malloc_hook = tr_mallochook;
        __realloc_hook = tr_reallochook;

        pthread_mutex_unlock (&_hook_lock);
    }

    bool oldval = __sync_val_compare_and_swap(&_in_trace, false, true);
    if (false == oldval)
    {
        assert(oldval == false && _in_trace == true);

        trace_where (caller);
        if (hdr == NULL) {
            /* Failed realloc.  */
            trace_what ("! %p %#lx\n", ptr, (unsigned long int) size);
        } else if (ptr == NULL) {
            trace_what ("+ %p %#lx\n", hdr, (unsigned long int) size);
        } else {
            trace_what ("< %p\n", ptr, 0/*ignored*/);
            trace_where (caller);
            trace_what ("> %p %#lx\n", hdr, (unsigned long int) size);
        }
        dump_stack();

        oldval = __sync_val_compare_and_swap(&_in_trace, true, false);
        assert(oldval == true && _in_trace == false);
    }

    if (hdr == mallwatch) {
        tr_break ();
    }

    return hdr;
}

static __ptr_t
tr_memalignhook (
     __malloc_size_t alignment, 
     __malloc_size_t size,
     const __ptr_t caller)
{
    __ptr_t hdr;

    {
        pthread_mutex_lock (&_hook_lock);

        __memalign_hook = _tr_old_memalign_hook;
        __malloc_hook = _tr_old_malloc_hook;
 
        if (_tr_old_memalign_hook != NULL) {
            hdr = (__ptr_t) (*_tr_old_memalign_hook) (alignment, size, caller);
        } else {
            hdr = (__ptr_t) memalign (alignment, size);
        }

        __memalign_hook = tr_memalignhook;
        __malloc_hook = tr_mallochook;

        pthread_mutex_unlock (&_hook_lock);
    }

    bool oldval = __sync_val_compare_and_swap(&_in_trace, false, true);
    if (false == oldval)
    {
        assert(oldval == false && _in_trace == true);

        trace_where (caller);
        /* We could be printing a NULL here; that's OK.  */
        trace_what ("+ %p %#lx\n", hdr, (unsigned long int) size);
        dump_stack();

        oldval = __sync_val_compare_and_swap(&_in_trace, true, false);
        assert(oldval == true && _in_trace == false);
    }

    if (hdr == mallwatch) {
        tr_break ();
    } 

    return hdr;
}



#ifdef _LIBC

/* This function gets called to make sure all memory the library
   allocates get freed and so does not irritate the user when studying
   the mtrace output.  */
static void __libc_freeres_fn_section
release_libc_mem (void)
{
    /* Only call the free function if we still are running in mtrace mode.  */
    if (_mallstream != NULL)
      __libc_freeres ();
}
#endif


/* We enable tracing if either the environment variable MALLOC_TRACE
   is set, or if the variable mallwatch has been patched to an address
   that the debugging user wants us to stop on.  When patching mallwatch,
   don't forget to set a breakpoint on tr_break!  */

extern "C" void
mtrace ()
{
    char *mallfile;

    //printf("mtrace() substitute\n");

    /* Don't panic if we're called more than once.  */
    if (_mallstream != NULL)
        return;

    /* Call backtrace before hooking because dynamically linked backtrace() will call malloc().
     * Otherwise we run in an infinite loop. 
     */
    void *stack_buf[STACK_DEPTH] = {0};
    int  n = mbacktrace(stack_buf, STACK_DEPTH);
    n = n; //Shut compiler warning

    mallfile = getenv (_mallenv);
    if (mallfile != NULL || mallwatch != NULL)
    {
        char *mtb = (char*) malloc (TRACE_ITEM_SIZE);
        if (mtb == NULL)
        return;

        _mallstream = fopen (mallfile != NULL ? mallfile : "/dev/null", "wc");
        if (_mallstream != NULL)
        {
            /* Make sure we close the file descriptor on exec.  */
            int flags = fcntl (fileno (_mallstream), F_GETFD, 0);
            if (flags >= 0)
            {
                flags |= FD_CLOEXEC;
                fcntl (fileno (_mallstream), F_SETFD, flags);
            }

            /* Be sure it doesn't malloc its buffer!  */
            _malloc_trace_buffer = mtb;
            setvbuf (_mallstream, _malloc_trace_buffer, _IOFBF, TRACE_ITEM_SIZE);

            fprintf (_mallstream, "= Start\n");

            _tr_old_free_hook = __free_hook;
            __free_hook = tr_freehook;
            _tr_old_malloc_hook = __malloc_hook;
            __malloc_hook = tr_mallochook;
            _tr_old_realloc_hook = __realloc_hook;
            __realloc_hook = tr_reallochook;
            _tr_old_memalign_hook = __memalign_hook;
            __memalign_hook = tr_memalignhook;
        }
        else
            free (mtb);
    }
}

extern "C" void
muntrace ()
{
    //printf("mUNtrace() substitute\n");

    if (_mallstream == NULL)
      return;

    fprintf (_mallstream, "= End\n");
    fclose (_mallstream);
    _mallstream = NULL;
    __free_hook = _tr_old_free_hook;
    __malloc_hook = _tr_old_malloc_hook;
    __realloc_hook = _tr_old_realloc_hook;
    __memalign_hook = _tr_old_memalign_hook;
}
